@@ -1,16 +1,22 @@
 // Copyright 2015 Google Inc. All rights reserved.
 //
-// Licensed under the Apache License, Version 2.0 (the "License");
-// you may not use this file except in compliance with the License.
-// You may obtain a copy of the License at
+// Permission is hereby granted, free of charge, to any person obtaining a copy
+// of this software and associated documentation files (the "Software"), to deal
+// in the Software without restriction, including without limitation the rights
+// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+// copies of the Software, and to permit persons to whom the Software is
+// furnished to do so, subject to the following conditions:
 //
-//     http://www.apache.org/licenses/LICENSE-2.0
+// The above copyright notice and this permission notice shall be included in
+// all copies or substantial portions of the Software.
 //
-// Unless required by applicable law or agreed to in writing, software
-// distributed under the License is distributed on an "AS IS" BASIS,
-// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-// See the License for the specific language governing permissions and
-// limitations under the License.
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+// THE SOFTWARE.
 
 //! Raw parser, for doing a single pass over input.
 
@@ -26,17 +32,21 @@ enum State {
     StartBlock,
     InContainers,
     Inline,
+    TableHead(usize, usize), // limit, next
+    TableBody,
+    TableRow,
     CodeLineStart,
     Code,
     InlineCode,
     Literal,
 }
 
-#[derive(Copy, Clone, Debug)]
+#[derive(Copy, Clone, Debug, PartialEq)]
 enum Container {
     BlockQuote,
     List(usize, u8),
     ListItem(usize),
+    FootnoteDefinition,
 }
 
 pub struct RawParser<'a> {
@@ -78,6 +88,13 @@ pub enum Tag<'a> {
     CodeBlock(Cow<'a, str>),
     List(Option<usize>),  // TODO: add delim and tight for ast (not needed for html)
     Item,
+    FootnoteDefinition(Cow<'a, str>),
+
+    // tables
+    Table(i32),
+    TableHead,
+    TableRow,
+    TableCell,
 
     // span-level tags
     Emphasis,
@@ -94,28 +111,21 @@ pub enum Event<'a> {
     Text(Cow<'a, str>),
     Html(Cow<'a, str>),
     InlineHtml(Cow<'a, str>),
+    FootnoteReference(Cow<'a, str>),
     SoftBreak,
     HardBreak,
 }
 
-pub struct Options(u32);
-
-const OPTION_FIRST_PASS: u32 = 1 << 0;
-
-const MAX_LINK_NEST: usize = 10;
-
-impl Options {
-    pub fn new() -> Options {
-        Options(0)
-    }
-    pub fn set_first_pass(&mut self) {
-        self.0 |= OPTION_FIRST_PASS;
-    }
-    pub fn is_first_pass(&self) -> bool {
-        (self.0 & OPTION_FIRST_PASS) != 0
+bitflags! {
+    flags Options: u32 {
+        const OPTION_FIRST_PASS = 1 << 0,
+        const OPTION_ENABLE_TABLES = 1 << 1,
+        const OPTION_ENABLE_FOOTNOTES = 1 << 1,
     }
 }
 
+const MAX_LINK_NEST: usize = 10;
+
 impl<'a> RawParser<'a> {
     pub fn new_with_links(text: &'a str, opts: Options,
             links: HashMap<String, (Cow<'a, str>, Cow<'a, str>)>) -> RawParser<'a> {
@@ -162,7 +172,7 @@ impl<'a> RawParser<'a> {
     }
 
     fn init_active(&mut self) {
-        if self.opts.is_first_pass() {
+        if self.opts.contains(OPTION_FIRST_PASS) {
             self.active_tab[b'\n' as usize] = 1
         } else {
             for &c in b"\x00\t\n\r_\\&*[!`<" {
@@ -188,19 +198,23 @@ impl<'a> RawParser<'a> {
         let (tag, _, next) = self.stack.pop().unwrap();
         match tag {
             // containers
-            Tag::BlockQuote | Tag::List(_) | Tag::Item => {
+            Tag::BlockQuote | Tag::List(_) | Tag::Item | Tag::FootnoteDefinition(_) => {
                 let _ = self.containers.pop();
             }
 
             // block level tags
-            Tag::Paragraph | Tag::Header(_) | Tag::Rule | Tag::CodeBlock(_) => {
+            Tag::Paragraph | Tag::Header(_) | Tag::Rule | Tag::CodeBlock(_) | Tag::Table(_) => {
                 self.state = State::StartBlock;
                 // TODO: skip blank lines (for cleaner source maps)
             }
 
+            // tables
+            Tag::TableCell => self.state = State::TableRow,
+            Tag::TableRow | Tag::TableHead => self.state = State::TableBody,
+
             // inline
             Tag::Code => self.state = State::Inline,
-            _ => ()
+            _ => (),
         }
         if next != 0 { self.off = next; }
 
@@ -249,6 +263,7 @@ impl<'a> RawParser<'a> {
                         return (i, false, space);
                     }
                 }
+                Container::FootnoteDefinition => (),
                 Container::List(_, _) => (),
                 Container::ListItem(indent) => {
                     if space >= indent {
@@ -310,30 +325,6 @@ impl<'a> RawParser<'a> {
         None
     }
 
-    // n is number of bytes (in blank lines) to skip
-    fn end_containing_lists(&mut self, n: usize) -> Event<'a> {
-        let mut i = self.stack.len();
-        while i >= 2 {
-            if let (Tag::List(_), _, _) = self.stack[i - 2] {
-                i -= 2;
-            } else {
-                break;
-            }
-        }
-        let mut next = self.off + n;
-        while i < self.stack.len() {
-            if let (Tag::List(start), _, _) = self.stack[i] {
-                self.stack[i] = (Tag::List(start), self.off, next);
-            }
-            if let (Tag::Item, _, _) = self.stack[i + 1] {
-                self.stack[i + 1] = (Tag::Item, self.off, self.off);
-            }
-            next = self.off;
-            i += 2;
-        }
-        self.end()
-    }
-
     fn start_block(&mut self) -> Option<Event<'a>> {
         let size = self.text.len();
         //println!("start_block {}", self.off);
@@ -356,11 +347,28 @@ impl<'a> RawParser<'a> {
             if at_eol {
                 self.off += n;
                 self.state = State::StartBlock;
-                // two empty lines close a list
+                // two empty lines closes lists and footnotes
                 let (n, empty_lines) = self.scan_empty_lines(&self.text[self.off ..]);
                 //println!("{} empty lines (n = {})", empty_lines, n);
-                if empty_lines >= 1 && self.at_list(2).is_some() {
-                    return Some(self.end_containing_lists(n));
+                let mut closed = false;
+                if empty_lines >= 1 {
+                    let mut close_tags: Vec<&mut (Tag<'a>, usize, usize)> = self.stack.iter_mut().skip_while(|tag| {
+                        match tag.0 {
+                            Tag::List(_) | Tag::FootnoteDefinition(_) => false,
+                            _ => true,
+                        }
+                    }).collect();
+                    if close_tags.len() != 0 {
+                        for tag in &mut close_tags {
+                            tag.1 = self.off; // limit
+                            tag.2 = self.off; // next
+                        }
+                        close_tags[0].2 = self.off + n; // next
+                        closed = true;
+                    }
+                }
+                if closed {
+                    return Some(self.end());
                 }
                 self.off += n;
                 if let Some(_) = self.at_list(2) {
@@ -449,6 +457,16 @@ impl<'a> RawParser<'a> {
                     }
                 }
                 b'[' => {
+                    if self.opts.contains(OPTION_ENABLE_FOOTNOTES) {
+                        if let Some((name, n)) = self.parse_footnote_definition(tail) {
+                            if self.containers.last() == Some(&Container::FootnoteDefinition) {
+                                return Some(self.end());
+                            }
+                            self.off += n;
+                            self.containers.push(Container::FootnoteDefinition);
+                            return Some(self.start(Tag::FootnoteDefinition(Cow::Borrowed(name)), self.text.len(), 0));
+                        }
+                    }
                     if self.try_link_reference_definition(tail) {
                         continue;
                     }
@@ -460,20 +478,33 @@ impl<'a> RawParser<'a> {
         None
     }
 
-    // can start a paragraph or a setext header, as they start similarly
+    // can start a paragraph, a setext header, or a table, as they start similarly
     fn start_paragraph(&mut self) -> Event<'a> {
         let mut i = self.off + scan_nextline(&self.text[self.off..]);
 
         if let (n, true, space) = self.scan_containers(&self.text[i..]) {
             i += n;
-            let (n, level) = scan_setext_header(&self.text[i..]);
-            if space <= 3 && n != 0 {
-                let next = i + n;
-                while i > self.off && is_ascii_whitespace(self.text.as_bytes()[i - 1]) {
-                    i -= 1;
+            if space <= 3 {
+                let (n, level) = scan_setext_header(&self.text[i..]);
+                if n != 0 {
+                    let next = i + n;
+                    while i > self.off && is_ascii_whitespace(self.text.as_bytes()[i - 1]) {
+                        i -= 1;
+                    }
+                    self.state = State::Inline;
+                    return self.start(Tag::Header(level), i, next);
+                }
+                if self.opts.contains(OPTION_ENABLE_TABLES) {
+                    let (n, cols) = scan_table_head(&self.text[i..]);
+                    if n != 0 {
+                        let next = i + n;
+                        while i > self.off && is_ascii_whitespace(self.text.as_bytes()[i - 1]) {
+                            i -= 1;
+                        }
+                        self.state = State::TableHead(i, next);
+                        return self.start(Tag::Table(cols), self.text.len(), 0);
+                    }
                 }
-                self.state = State::Inline;
-                return self.start(Tag::Header(level), i, next);
             }
         }
 
@@ -482,6 +513,34 @@ impl<'a> RawParser<'a> {
         self.start(Tag::Paragraph, size, 0)
     }
 
+    fn start_table_head(&mut self) -> Event<'a> {
+        assert!(self.opts.contains(OPTION_ENABLE_TABLES));
+        if let State::TableHead(limit, next) = self.state {
+            self.state = State::TableRow;
+            return self.start(Tag::TableHead, limit, next);
+        } else {
+            panic!();
+        }
+    }
+
+    fn start_table_body(&mut self) -> Event<'a> {
+        assert!(self.opts.contains(OPTION_ENABLE_TABLES));
+        let (off, _) = match self.scan_containers(&self.text[self.off ..]) {
+            (n, true, space) => (self.off + n, space),
+            _ => {
+                return self.end();
+            }
+        };
+        let n = scan_blank_line(&self.text[off..]);
+        if n != 0 {
+            self.off = off + n;
+            return self.end();
+        }
+        self.state = State::TableRow;
+        self.off = off;
+        return self.start(Tag::TableRow, self.text.len(), 0);
+    }
+
     fn start_hrule(&mut self) -> Event<'a> {
         let limit = self.off;  // body of hrule is empty
         self.state = State::Inline;  // handy state for producing correct end tag
@@ -804,6 +863,49 @@ impl<'a> RawParser<'a> {
                     self.is_html_block(data))
     }
 
+    fn next_table_cell(&mut self) -> Event<'a> {
+        assert!(self.opts.contains(OPTION_ENABLE_TABLES));
+        let bytes   = self.text.as_bytes();
+        let mut beg = self.off + scan_whitespace_no_nl(&self.text[self.off ..]);
+        let mut i   = beg;
+        let limit   = self.limit();
+        if i < limit && bytes[i] == b'|' {
+            i   += 1;
+            beg += 1;
+            self.off += 1;
+        }
+        if i >= limit {
+            self.off = limit;
+            return self.end();
+        }
+        let mut n = 0;
+        while i < limit {
+            let c = bytes[i];
+            if c == b'\\' && i + 1 < limit && bytes[i + 1] == b'|' {
+                i += 2;
+                continue;
+            } else if c == b'|' {
+                n = 1;
+                break;
+            }
+            n = scan_blank_line(&self.text[i..]);
+            if n != 0 {
+                if i > beg {
+                    n = 0;
+                }
+                break;
+            }
+            i += 1;
+        }
+        if i > beg {
+            self.state = State::Inline;
+            self.start(Tag::TableCell, i, i + n)
+        } else {
+            self.off = i + n;
+            self.end()
+        }
+    }
+
     fn next_inline(&mut self) -> Event<'a> {
         let bytes = self.text.as_bytes();
         let beg = self.off;
@@ -863,6 +965,7 @@ impl<'a> RawParser<'a> {
             b'&' => self.char_entity(),
             b'_' => self.char_emphasis(),
             b'*' => self.char_emphasis(),
+            b'[' if self.opts.contains(OPTION_ENABLE_FOOTNOTES) => self.char_link_footnote(),
             b'[' | b'!' => self.char_link(),
             b'`' => self.char_backtick(),
             b'<' => self.char_lt(),
@@ -977,6 +1080,12 @@ impl<'a> RawParser<'a> {
                     i += 1;
                 }
             } else if c2 == b'[' {
+                if self.opts.contains(OPTION_ENABLE_FOOTNOTES) {
+                    if let Some((_, n)) = self.parse_footnote(&self.text[i..limit]) {
+                        i += n;
+                        continue;
+                    }
+                }
                 if let Some((_, _, _, n)) = self.parse_link(&self.text[i..limit], false) {
                     i += n;
                 } else {
@@ -1173,6 +1282,9 @@ impl<'a> RawParser<'a> {
                     }
                 }
                 b'[' => {
+                    if self.opts.contains(OPTION_ENABLE_FOOTNOTES) && self.parse_footnote(&data[i..]).is_some() {
+                        return false;
+                    }
                     if self.parse_link(&data[i..], true).is_some() { return true; }
                 }
                 b'\\' => i += 1,
@@ -1199,6 +1311,65 @@ impl<'a> RawParser<'a> {
         false
     }
 
+    // # Footnotes
+
+    fn parse_footnote_definition<'b>(&self, data: &'b str) -> Option<(&'b str, usize)> {
+        assert!(self.opts.contains(OPTION_ENABLE_FOOTNOTES));
+        self.parse_footnote(data).and_then(|(name, len)| {
+            let n_colon = scan_ch(&data[len ..], b':');
+            if n_colon == 0 {
+                None
+            } else {
+                let space = scan_whitespace_no_nl(&data[len + n_colon..]);
+                Some((name, len + n_colon + space))
+            }
+        })
+    }
+
+    fn char_link_footnote(&mut self) -> Option<Event<'a>> {
+        assert!(self.opts.contains(OPTION_ENABLE_FOOTNOTES));
+        if let Some((name, end)) = self.parse_footnote(&self.text[self.off .. self.limit()]) {
+            self.off += end;
+            Some(Event::FootnoteReference(Cow::Borrowed(name)))
+        } else {
+            self.char_link()
+        }
+    }
+
+    fn parse_footnote<'b>(&self, data: &'b str) -> Option<(&'b str, usize)> {
+        assert!(self.opts.contains(OPTION_ENABLE_FOOTNOTES));
+        let (n_footnote, text_beg, text_end) = self.scan_footnote_label(data);
+        if n_footnote == 0 { return None; }
+        return Some((&data[text_beg..text_end], n_footnote));
+    }
+
+    fn scan_footnote_label(&self, data: &str) -> (usize, usize, usize) {
+        assert!(self.opts.contains(OPTION_ENABLE_FOOTNOTES));
+        let mut i = scan_ch(data, b'[');
+        if i == 0 { return (0, 0, 0); }
+        if i >= data.len() || data.as_bytes()[i] != b'^' { return (0, 0, 0); }
+        i += 1;
+        let text_beg = i;
+        loop {
+            if i >= data.len() { return (0, 0, 0); }
+            match data.as_bytes()[i] {
+                b'\n' => {
+                    let n = self.scan_whitespace_inline(&data[i..]);
+                    if n == 0 { return (0, 0, 0); }
+                    i += n;
+                    continue;
+                }
+                b']' => break,
+                b'\\' => i += 1,
+                _ => ()
+            }
+            i += 1;
+        }
+        let text_end = i;
+        i += 1;  // skip closing ]
+        (i, text_beg, text_end)
+    }
+
     // # Autolinks and inline HTML
 
     fn char_lt(&mut self) -> Option<Event<'a>> {
@@ -1487,6 +1658,9 @@ impl<'a> Iterator for RawParser<'a> {
                     }
                 }
                 State::Inline => return Some(self.next_inline()),
+                State::TableHead(_, _) => return Some(self.start_table_head()),
+                State::TableBody => return Some(self.start_table_body()),
+                State::TableRow => return Some(self.next_table_cell()),
                 State::CodeLineStart => return Some(self.next_code_line_start()),
                 State::Code => return Some(self.next_code()),
                 State::InlineCode => return Some(self.next_inline_code()),
